#include <iostream>
#include <fstream>
#include <math.h>

using namespace std;


float fmod(float A, float B) {
	return A*float(int(A/B)%1);
}

int imax(int A, int B) {
	return (A>B)?A:B;
}
int imin(int A, int B) {
	return (A<B)?A:B;
}
float min(float A, float B) {
	return (A<B)?A:B;
}
float max(float A, float B) {
	return (A>B)?A:B;
}

template <class T>
T abs(T A) {
	return (A>0)?A:(-A);
}

void swap(int& A, int& B) {
	int C = A;
	A = B;
	B = C;
}

double arctan(double x, double y) { //use x,y coords to find angle from 0 to 2*pi instead of the usual -pi/2 to pi/2
	if (x>0 && y>0) return atan(y/x);
	else if (x==0 && y>0) return 3.14159265358979/2.0;
	else if (x==0 && y<0) return 3.14159265358979*3.0/2.0;
	else if (x==0 && y==0) return 0.0;
	else if (x<0) return 3.14159265358979+atan(y/x);
	else if (y<0) return atan(y/x)+2.0*3.14159265358979;
}

class window { //don't worry.  No GUI support.  Just outputs to a nice ppm file.
private:
	int idim;
	int jdim;
	float xmin, ymin, xmax, ymax;
	char* pixels;
public:
	window(int isize, int jsize) { //initialize window with pixel dimensions
		idim = isize;
		jdim = jsize;
		pixels = new char[isize*jsize];
		for(int i=0; i<isize*jsize; i++) {
			pixels[i] = 0;
		}
	}
	~window() {
		delete [] pixels;
	}
	void xfrom(float xmn, float xmx) { //set x-bounds
		xmin = xmn;
		xmax = xmx;
	}
	void yfrom(float ymn, float ymx) { //set y-bounds
		ymin = ymn;
		ymax = ymx;
	}
	int isize() {
		return idim;
	}
	int jsize() {
		return jdim;
	}
	void clear() {
		for(int i=0; i<idim*jdim; i++) {
			pixels[i] = 0;
		}
	}
	void pxOn(int i, int j, int col) { //turn on a pixel with integer coords and color index
		if (i>0 && i<idim && j>0 && j<jdim)
			pixels[i+idim*j] = col;
	}
	void pxOff(int i, int j) { //turn off a pixel (black)
		if (i>0 && i<idim && j>0 && j<jdim)
			pixels[i+idim*j] = 0;
	}
	void pxChg(int i, int j) { //color 1 if 0, 0 if 1
		if (i>0 && i<idim && j>0 && j<jdim)
			if (pixels[i+idim*j] == 0) {
				pixels[i+idim*j] = 1;
			} else {
				pixels[i+idim*j] = 0;
			}
	}
	void ptOn(float x, float y, int col) { //turn on a point with x,y coords
		if (x>xmin && x<xmax && y>ymin && y<ymax)
			pxOn(int((x-xmin)/(xmax-xmin)*idim),
			     int((y-ymin)/(ymax-ymin)*jdim), col);
	}
	void ptOff(float x, float y) { //turn off a point
		if (x>xmin && x<xmax && y>ymin && y<ymax)
			pxOff(int((x-xmin)/(xmax-xmin)*idim),
			      int((y-ymin)/(ymax-ymin)*jdim));
	}
	void ptChg(float x, float y) { //color 1 if 0, 0 if 1
		if (x>xmin && x<xmax && y>ymin && y<ymax)
			pxChg(int((x-xmin)/(xmax-xmin)*idim),
			      int((y-ymin)/(ymax-ymin)*jdim));
	}

	void ppm_out(char fname[]) { //output to a unix ppm image file
		ofstream output (fname);
		output << "P3\n#.\n" << idim << " " << jdim << "\n255\n";
		for(int j=jdim-1; j>=0; j--) {
			for(int i=0; i<idim; i++) {
				if(pixels[i+idim*j] == 0) {
					output << "0 0 0 ";
				} else if(pixels[i+idim*j] == 1) {
					output << "255 255 255 ";
				} else if(pixels[i+idim*j] == 2) {
					output << "255 0 0 ";
				} else {
					output << "128 128 128 ";
				}
			}
		}
		output.close();
	}
	void ptLine(float x0, float y0, float x1, float y1, int col) { //draw a line from (x0,y0), (x1,y1)
		
		pxLine(
			int((x0-xmin)/(xmax-xmin)*idim),
			int((y0-ymin)/(ymax-ymin)*jdim),
			int((x1-xmin)/(xmax-xmin)*idim),
			int((y1-ymin)/(ymax-ymin)*jdim), col
		);
	}
	void pxLine(int x0, int y0, int x1, int y1, int col) { //draw line between to pixels
	if (x0 != x1 && y0 != y1) {
		bool steep = abs(y1-y0) > abs(x1-x0);
		if(steep) {
			swap(x0, y0);
			swap(x1, y1);
		}
		if (x0 > x1) {
			swap(x0,x1);
			swap(y0,y1);
		}
		int dx = abs(x1-x0);
		int dy = abs(y1-y0);
		int err = 0;
		int derr = dy;
		int py = y0;
		int ystep = -1;
		if (y0 < y1) {
			ystep = 1;
		}
		for(int px=x0; px<=x1; px++) {
			if (steep) {
				pxOn(py,px, col);
			} else {
				pxOn(px,py, col);
			}
			err += derr;
			if (2*err >= dx) {
				py += ystep;
				err -= dx;
			}
		}
	}
	}
};

int main(int argc, char* argv []){
	FILE* fptr;
	fptr = fopen(argv[1],"r");
	char letter1, letter2;
	char fstring [64];
	char objname [64];

	int lines = 0;
	while(!feof(fptr)) {
		if(getc(fptr) == 13) lines++;
	}
	cout << lines << " lines." << endl;
	
	//ppm settings
	double pi = 3.14159265358979;
	window uv_window (1024,768);
	uv_window.xfrom(0,1);
	uv_window.yfrom(0,1);

	int nvertices = 0;	
	int nlines=0;
	int nnormals = 0;
	int nfaces = 0;
	short int buffer[10];
	int c=0;
	rewind(fptr);
	float xp, yp, zp;
	
	while(nlines < lines) {
		nlines++;
		letter1 = getc(fptr);
		switch(letter1) {
		case '#':
			letter2 = '\0';
			while(letter2 != '\n') {
				letter2 = getc(fptr);
			}
		break;
		case 'm':
			fscanf(fptr,"%s",fstring);
			letter2 = '\0';
			while(letter2 != '\n') {
				letter2 = getc(fptr);
			}
		break;
		case 'o':
			fscanf(fptr,"%s",objname);
			cout << objname;
			letter2 = '\0';
			while(letter2 != '\n') {
				letter2 = getc(fptr);
			}
		break;
		case 'v':
			switch(getc(fptr)) {
			case ' ':
				fscanf(fptr, "%f %f %f", &xp, &yp, &zp);
				nvertices++;
				getc(fptr);
				getc(fptr);
			break;
			case 'n':
				nnormals++;
				fscanf(fptr, "%f %f %f", &xp, &yp, &zp);
				getc(fptr);
				getc(fptr);
			break;
			case 't':
				fscanf(fptr, "%f %f %f", &xp, &yp, &zp);
				getc(fptr);
				getc(fptr);
			break;
			//default: cout << "weird..." << endl;
			}
		break;
		case 'f':
			nfaces++;
			getc(fptr);
			for(int j=0; j<3; j++) {
				c=0;
				buffer[c]=getc(fptr);
				while(buffer[c] != '/') {
					c++;
					buffer[c] = getc(fptr);
				}
				getc(fptr);
				c=0;
				buffer[c]=getc(fptr);
				while(buffer[c] != ' ' && buffer[c] != 13) {
					c++;
					buffer[c] = getc(fptr);
				}
			}
			//cout << endl;
			getc(fptr);
		break;
		case 'g':
			fscanf(fptr,"%s",fstring);
			letter2 = '\0';
			while(letter2 != '\n') {
				letter2 = getc(fptr);
			}
		break;
		case 'u':
			fscanf(fptr,"%s",fstring);
			letter2 = '\0';
			while(letter2 != '\n') {
				letter2 = getc(fptr);
			}
		break;
		default:
			cout << "line " << nlines << ": unknown character, "<<0+letter1<<"." << endl;//if you get this, you probably
			//didn't triangulate the mesh in wings, or the file doesn't match the expected format well enough.
			letter2 = '\0';
			while(letter2 != '\n') {
				letter2 = getc(fptr);
			}
		}
	}
	cout << nvertices << " vertices." << endl;
	cout << nnormals << " normal vectors." << endl;
	cout << nfaces << " faces." << endl;


	float xvertices [nvertices];
	float yvertices [nvertices];
	float zvertices [nvertices];

	float xnormals [nvertices];
	float ynormals [nnormals];
	float znormals [nnormals];
	
	int vindices [nfaces][3];
	int nindices [nfaces][3];
	int uvindices [nfaces][3];
	
	float u[nvertices];
	float v[nvertices];


	//Now that we know what the file looks like, create the arrays (above) and then rewind and read the data
	//into the arrays.  Just checks it for key letters at the beginning of the lines and inserts them into the arrays
	int cvertex = 0;
	int cnormal = 0;
	int cface = 0;
	rewind(fptr);
	nlines = 0;
	while(nlines < lines) {
		nlines++;
		letter1 = getc(fptr);
		switch(letter1) {
		case '#':
			letter2 = '\0';
			while(letter2 != '\n') {
				letter2 = getc(fptr);
			}
		break;
		case 'm':
			fscanf(fptr,"%s",fstring);
			letter2 = '\0';
			while(letter2 != '\n') {
				letter2 = getc(fptr);
			}
		break;
		case 'o':
			fscanf(fptr,"%s",fstring);
			letter2 = '\0';
			while(letter2 != '\n') {
				letter2 = getc(fptr);
			}
		break;
		case 'v':
			switch(getc(fptr)) {
			case ' ':
				fscanf(fptr, "%f %f %f", &xp, &yp, &zp);
				getc(fptr);
				getc(fptr);
				xvertices[cvertex] = xp;
				yvertices[cvertex] = yp;
				zvertices[cvertex] = zp;
				cvertex++;
			break;
			case 'n':
				fscanf(fptr, "%f %f %f", &xp, &yp, &zp);
				getc(fptr);
				getc(fptr);
				xnormals[cnormal] = xp;
				ynormals[cnormal] = yp;
				znormals[cnormal] = zp;
				cnormal++;
			break;
			case 't':
				fscanf(fptr, "%f %f %f", &xp, &yp, &zp);
				getc(fptr);
				getc(fptr);
			break;
			//default: cout << "weird..." << endl;
			}
		break;
		case 'f':
			getc(fptr);
			for(int j=0; j<3; j++) {
				c=0;
				buffer[c]=getc(fptr);
				while(buffer[c] != '/') {
					c++;
					buffer[c] = getc(fptr);
				}
				vindices[cface][j] = 0;
				for(int i=0; i<c; i++) { //convert ascii[] to dec
					vindices[cface][j] += (buffer[i]-48)*int(pow(10.0,c-i-1));
				}
				vindices[cface][j]--;
				getc(fptr);
				c=0;
				buffer[c]=getc(fptr);
				while(buffer[c] != ' ' && buffer[c] != 13) {
					c++;
					buffer[c] = getc(fptr);
				}
				nindices[cface][j] = 0;
				for(int i=0; i<c; i++) { //convert ascii[] to dec
					nindices[cface][j] += (buffer[i]-48)*int(pow(10.0,c-i-1));
				}
				nindices[cface][j]--;
			}
			cface++;
			getc(fptr);
		break;
		case 'g':
			fscanf(fptr,"%s",fstring);
			letter2 = '\0';
			while(letter2 != '\n') {
				letter2 = getc(fptr);
			}
		break;
		case 'u':
			fscanf(fptr,"%s",fstring);
			letter2 = '\0';
			while(letter2 != '\n') {
				letter2 = getc(fptr);
			}
		break;
		default:
			letter2 = '\0';
			while(letter2 != '\n') {
				letter2 = getc(fptr);
			}
		}
	}
/*	
	//check for connectivity: not completed yet but will allow cloth-like simulations for 
	//better uv-mapping in the future - perhaps the distant future.
	int max_connections = 0;
	int connect_count [nvertices];
	for(int i=0; i<nvertices; i++) {
		connect_count[i] = 0;
	}
	for(int i=0; i<nfaces; i++) {
		max_connections = imax(max_connections, (connect_count[vindices[i][0]]++)); //tally up each vertex's
		max_connections = imax(max_connections, (connect_count[vindices[i][1]]++)); //connections with other
		max_connections = imax(max_connections, (connect_count[vindices[i][2]]++)); //vertices - each counted twice
		// as long as it belongs to two adjacent faces - so as long as the mesh is closed.
	}
	max_connections = -int(-double(max_connections)/2.0);
	cout << "max connections: " << max_connections << endl;
*/
	
	//calculate uv vectors - spherical mapping

	//SET UP SPHERICAL COORDINATE SYSTEM
	float x_offset = 1e-5;
	float y_offset = 1.0;  //spherical mapping centered at <0,1,0>
	float z_offset = 1e-5; //use small number instead of 0.  No good reason, but it works this way.
	float x_rotate = 90.0; //rotated 90 degrees.
	float y_rotate = 0.0;
	float z_rotate = 0.0;
	
	x_rotate *= pi/180.0;
	y_rotate *= pi/180.0;
	z_rotate *= pi/180.0;
	float r, theta, phi;
	float x,y,z, x2, y2, z2;
	for(int i=0; i<nvertices; i++) {
		x2 = xvertices[i] - x_offset;
		y2 = yvertices[i] - y_offset;
		z2 = zvertices[i] - z_offset;
		x = x2;
		y = y2*cos(x_rotate)-z2*sin(x_rotate);
		z = z2*cos(x_rotate)+y2*sin(x_rotate);
		x2 = x*cos(y_rotate)+z*sin(y_rotate);
		y2 = y;
		z2 = z*cos(y_rotate)-x*sin(y_rotate);
		x = x2*cos(z_rotate)-y2*sin(z_rotate);
		y = y2*cos(z_rotate)+x2*sin(z_rotate);
		z = z2;
		r = sqrt(x*x+y*y);
		theta = arctan(y,x);
		phi = atan(z/r);
		u[i] = theta/2.0/pi;
		v[i] = (phi+pi/2.0)/pi;
		uv_window.ptOn(u[i],v[i], 1);
	}

	//now that we have the uv-vectors, check to see if shifting them accross the discontinuity at
	//theta=2*pi to theta=0 will decrease the area of the uv system faces.  If it does help, then
	//tally it up and come back later once the proper amount of space to store all extra uv-vectors
	//has been allocated.
	//***this has been changed: now it just checkes the horizontal width of the face rather than the 
	//actual area of the face.  It just gives better results.  No other reason really

	bool toMove[nfaces][3];
	for(int i=0; i<nfaces; i++) {
		toMove[i][0] = toMove[i][1] = toMove[i][2] = 0;
	}

	float u1, u2, u3, v1, v2, v3;
	float u1n, u2n, u3n;
	float area[4];
	short int iarea;

	for(int i=0; i<nfaces; i++) {
		//check moving point 1 of 3.  If it is on the right side of [0,1]x[0,1], move it all
		//the way over past the left side and vice versa.
		u1 = u[vindices[i][0]];
		u2 = u[vindices[i][1]];
		u3 = u[vindices[i][2]];
		v1 = v[vindices[i][0]];
		v2 = v[vindices[i][1]];
		v3 = v[vindices[i][2]];
		u1n = u[vindices[i][0]];
		if (u1n>0.5) {
			u1n -= 1.0;
		} else {
			u1n += 1.0;
		}
		u2n = u[vindices[i][1]];
		if (u2n>0.5) {
			u2n -= 1.0;
		} else {
			u2n += 1.0;
		}
		u3n = u[vindices[i][2]];
		if (u3n>0.5) {
			u3n -= 1.0;
		} else {
			u3n += 1.0;
		}
		//area[0] = abs((u3-u1)*(v2-v1)-(v3-v1)*(u2-u1))/2.0; //calculate areas
		//area[1] = abs((u3-u1n)*(v2-v1)-(v3-v1)*(u2-u1n))/2.0;
		//area[2] = abs((u3-u1)*(v2-v1)-(v3-v1)*(u2n-u1))/2.0;
		//area[3] = abs((u3n-u1)*(v2-v1)-(v3-v1)*(u2-u1))/2.0;
		area[0] = max(max(abs(u2-u1),abs(u3-u2)),abs(u3-u1)); //calculate spacing of u-coordinates instead
		area[1] = max(max(abs(u2-u1n),abs(u3-u2)),abs(u3-u1n));
		area[2] = max(max(abs(u2n-u1),abs(u3-u2n)),abs(u3-u1));
		area[3] = max(max(abs(u2-u1),abs(u3n-u2)),abs(u3n-u1));

		if (area[1]<area[2] && area[1]<area[3]) iarea = 0;
		if (area[2]<area[1] && area[2]<area[3]) iarea = 1;
		if (area[3]<area[1] && area[3]<area[2]) iarea = 2;

		if (area[iarea+1] < area[0]) {
			toMove[i][iarea] = 1;
		}
	}
	
	//Draw lines in the (non-existent) window.  Good for painting a texture

	int i1, i2, i3;
	for(int i=0; i<nfaces; i++) {
		i1 = vindices[i][0];
		i2 = vindices[i][1];
		i3 = vindices[i][2];
		if(toMove[i][0]==0 && toMove[i][1]==0 && toMove[i][2]==0) { //if the face is good, outline it with lines
			uv_window.ptLine(u[i1],v[i1],u[i2],v[i2], 1);
			uv_window.ptLine(u[i1],v[i1],u[i3],v[i3], 1);
			uv_window.ptLine(u[i2],v[i2],u[i3],v[i3], 1);
		} else {
			uv_window.ptLine(u[i1],v[i1],u[i2],v[i2], 2); //if the face is no good, use red lines
			uv_window.ptLine(u[i1],v[i1],u[i3],v[i3], 2);
			uv_window.ptLine(u[i2],v[i2],u[i3],v[i3], 2);
		}
	}
	
	
	
	uv_window.ppm_out("uvs.ppm");
	system("open uvs.ppm");//comment line to prevent opening of ppm file




	//write a pov-ray mesh2 file from the data.
	char outname [64];
	char letter = 1;
	int i=0;
	strcpy(outname,objname);
	strcat(outname,".inc");
	cout << "file out = "<< outname << endl;
	ofstream povout (outname);
	povout << "vertex_vectors{ "<< nvertices << ",\n";
	for(int i=0; i<nvertices; i++) {
		povout << "<"<< xvertices[i] << "," << yvertices[i] << "," << zvertices[i] << ">";
		if (i<nvertices-1) {
			povout << ",\n";
		} else {
			povout << "\n";
		}
	}
	povout << "}\nnormal_vectors{ " << nnormals << ",\n";
	for(int i=0; i<nnormals; i++) {
		povout << "<"<< xnormals[i] << "," << ynormals[i] << "," << znormals[i] << ">";
		if (i<nnormals-1) {
			povout << ",\n";
		} else {
			povout << "\n";
		}
	}
	povout << "}\nuv_vectors{ " << nvertices << ",\n";
	for(int i=0; i<nvertices; i++) {
		povout << "<"<<u[i]<<","<<v[i]<<">";
		if (i<nvertices-1) {
			povout << ",\n";
		} else {
			povout << "\n";
		}
	}

	povout << "}\nface_indices{ "<<nfaces<<",\n";
	for(int i=0; i<nfaces; i++) {
		povout << "<"<<vindices[i][0]<<","<<vindices[i][1]<<","<<vindices[i][2]<<">";
		if (i<nfaces-1) {
			povout << ",\n";
		} else {
			povout << "\n";
		}
	}

	povout << "}\nnormal_indices{ "<<nfaces<<",\n";
	for(int i=0; i<nfaces; i++) {
		povout << "<"<<nindices[i][0]<<","<<nindices[i][1]<<","<<nindices[i][2]<<">";
		if (i<nfaces-1) {
			povout << ",\n";
		} else {
			povout << "\n";
		}
	}
	povout << "}\nuv_indices{ "<<nfaces<<",\n";
	for(int i=0; i<nfaces; i++) {
		povout << "<"<<vindices[i][0]<<","<<vindices[i][1]<<","<<vindices[i][2]<<">";
		if (i<nfaces-1) {
			povout << ",\n";
		} else {
			povout << "\n";
		}
	}
	povout << "}\n";
	povout.close();

	return 0;
}

